#!/usr/bin/env bash

# There are two ways to run this script:
# 1) Manually by running run_test.sh.
#    This can be done in your conda environment or MicadoWISE docker image.
# 2) Automatically through conda-build.
#    This is done in the CI or through `conda build .`.
#    Running the script this way requires pip-installation of packages that
#    do not have a conda package yet.
#
# There are several combinations of optional external dependencies that can be
# used.
# a) with or without 'common',
# b) with or without a PostgreSQL database, requires common, or
# c) with or without CPL Recipes included.
#
# Currently, setting MICADO_INTEGRATION_TESTS=True will test them all,
# otherwise only functional tests (and the notebooks) are performed.
#
# Optional dependencies can be pre-installed manually (see README), or
# automatically, e.g. in conda-build or through the Continous Integration.
# Automatic installation requires the following environment variables to be set:
# - common and micadopipeline: CONDA_DOWNLOAD_KEY (ask Gijs Verdoes Kleijn),
# - database: POSTGRES_DB, POSTGRES_USER, POSTGRES_PASSWORD, administrator
#   password to database, see README or .gitlab-ci.yml

# TODO: Ensure ScopeSIM is optional too.

# Echo commands and exit on first error
set -xe

# Everything in braces to synchronize stderr and stdout.
{

# https://stackoverflow.com/questions/5014823/how-to-profile-a-bash-shell-script-slow-startup
PS4='+ $EPOCHREALTIME\011 '

# Set SRC_DIR if necessary, e.g. when running outside CI.
SRC_DIR="${SRC_DIR-$(pwd)}"

# TODO: Fix coverage. Two problems:
# 1) run_test.sh can also be ran outside conda-build, so --source should be
#    different then.
# 2) The coverage of several jobs needs to be combined in the CI.
coverage erase



##################################
# Install optional dependencies. #
##################################

# Running the tests through conda-build requires pip-installation of several
# packages.
# TODO: Ensure MicadoWISE works without these as well.
python -c "import skycalc_ipy" || pip install skycalc_ipy
python -c "import scopesim" || pip install "ScopeSIM==0.1.4"
# ScopeSim_Templates fetches data from the internet, which does not work
# when using conda-build.
python -c "import scopesim_templates" || pip install git+https://github.com/astronomyk/ScopeSim_Templates.git@hb/micadowise
# TODO: Fetch ScopeSim data package.


if [[ -n ${MICADO_INTEGRATION_TESTS:+x} ]]; then
    echo "MICADO_INTEGRATION_TESTS set, installing optional dependencies."

    # Install common if necessary.
    if [[ -n "${CONDA_DOWNLOAD_KEY+x}" ]]; then
        # Ensure astro-wise.org channel has low priority, but default even
        # lower.
        conda config --append channels https://${CONDA_DOWNLOAD_KEY}@conda.astro-wise.org
        conda config --append channels defaults
        # Apparently mamba cannot install these packages from an
        # authenticated channel.
        python -c "import common" || conda install -y common micadopipeline
    fi

    # Directory where 'micado' is imported from, to check the coverage in.
    # This is either in the conda site packages directory, when running through
    # conda build, e.g.
    # /home/user/miniconda3/envs/micadowise/lib/python3.9/site-packages/micado
    # or the local checkout of the code, when running simple_test, e.g.
    # /home/user/micadowise/src/micado or
    # /opt/project/src/micado
    MICADO_DIR="$(python -c 'import micado; print(micado.__path__[0])')"

    # Set database credentials, if not set already.
    if [[ -n ${GITLAB_CI+x} ]]; then
        export \
            database_tablespacename="pg_default" \
            database_engine="postgresql" \
            database_port="5432" \
            database_name="postgres/${POSTGRES_DB}" \
            database_user="${POSTGRES_USER}" \
            database_password="${POSTGRES_PASSWORD}"

        # TODO: Move this such that it is easy to setup the database
        #       in docker-compose as well.
        if [[ ${database_engine:-filebased} != filebased ]]; then
            # Create database structures for all persistent classes.
            env \
                ask_administrator_password='' \
            coverage run --branch --append --source="${MICADO_DIR}" -m \
                micado.toolbox.backends.postgresql.dbtestsetup
        fi

        export \
            database_user="AWTEST" \
            database_password="lmno"
    fi
fi



####################
# Start the tests. #
####################

# Run functional tests. These should always succeed and should not require
# any external dependencies except ScopeSIM.
echo "Running functional tests."
coverage run --branch --source="${MICADO_DIR}" -m pytest \
    "${SRC_DIR}/src/micado/test/functional"

# Do integration tests, in particular database and CPL tests.
# These tests are not required to build the package, only in CI.
if [[ -n ${MICADO_INTEGRATION_TESTS:+x} ]]; then
    echo "MICADO_INTEGRATION_TESTS set, performing integration tests."

    # Run tests that actually test the database.
    if [[ ${database_engine:-filebased} != filebased ]]; then
        echo "Run database tests."
        coverage run --branch --append --source="${MICADO_DIR}" -m pytest \
            "${SRC_DIR}/src/micado/test/database"
    else
        echo "Skipping database tests."
    fi

    coverage run --branch --append --source="${MICADO_DIR}" -m pytest \
        "${SRC_DIR}/src/micado/test/integration"
else
    echo "MICADO_INTEGRATION_TESTS unset, skipping integration tests."
fi

echo "Running notebooks."
# TODO: Figure out a way to run all notebooks automatically.
#       Doing *.py will not always result in a positive exit code.
coverage run --branch --append --source="${MICADO_DIR}" "${SRC_DIR}"/notebooks/DigitalDesignGuide.py
coverage run --branch --append --source="${MICADO_DIR}" "${SRC_DIR}"/notebooks/SimulateRawScience.py
coverage run --branch --append --source="${MICADO_DIR}" "${SRC_DIR}"/notebooks/SimulateFlat.py
coverage run --branch --append --source="${MICADO_DIR}" "${SRC_DIR}"/notebooks/FullStandardImaging.py
coverage run --branch --append --source="${MICADO_DIR}" "${SRC_DIR}"/notebooks/RecipeValidation.py

# Generate coverage report.
coverage report -m



###############
# Check Style #
###############

# Ignore E203 because black disagrees.
# https://black.readthedocs.io/en/stable/the_black_code_style.html#slices
# TODO: Reenable flake8
#flake8 --docstring-convention google --max-line-length 120 --max-doc-length 80 --extend-ignore=E203 "${SRC_DIR}/src/micado"

black --check "${SRC_DIR}/src/micado"

# Install sourcery. This might break things and is thus disabled for now.
#pip install sourcery-cli
# Login if SOURCERY_TOKEN (a masked CI variable) is set.
#if [[ -n ${SOURCERY_TOKEN+x} ]]; then
#    sourcery login --token "${SOURCERY_TOKEN}"
#fi
# Find changed files and pass them to Sourcery
# git diff master --name-only | xargs sourcery refactor --check
# For now, just run on the whole code base.
# Sourcery is disabled for now, because it makes it impossible for
# others to run conda-build.
# TODO: Re-enable sourcery for CI only.
# sourcery refactor --check "${SRC_DIR}/src/micado"

} 2>&1
