#!/usr/bin/env bash

# https://betterdev.blog/minimal-safe-bash-script-template/
set -Eeuo pipefail

script_dir=$(cd "$(dirname "${BASH_SOURCE[0]}")" &>/dev/null && pwd -P)
cd "${script_dir}/../.."

conda init bash
source "${HOME}/.bashrc"
conda create -n micadowise
conda activate micadowise

# Build and install MicadoWISE from the repository. This makes it possible to
# create the container before publishing the MicadoWISE package.
conda mambabuild .
mamba install -c "${HOME}/conda-bld" micado

# Also install pip dependencies.
# Directly installing from micadowise.yml does not work because that will
# fetch the micadowise package from the omegacen channel again, overruling the
# package just built and installed.
# mamba env update --file micadowise.yml

# MicadoWISE does not actually depend on wcslib, but micadopipeline
# does, but that needs 7.3, which it does not have as a dependencies.
# TODO: move dependencies to micadopipeline.
# Specific versions of openjdk and jupytext are required for build.sh
# to behave consistently. For some reason these are not always properly
# selected when installing the MicadoWISE package even though they are
# defined in meta.yaml.
mamba install "wcslib >=7" "openjdk==11.0.9.1" "jupytext==1.13.1"
# ScopeSIM 0.2.0+ breaks simulating a cluster for large detectors.
# TODO: Upgrade to ScopeSIM 0.4.1+.
pip install skycalc_ipy "ScopeSim==0.1.4" ScopeSim_Templates

# Cleaning everything removes about a gigabyte from the Docker image.
conda clean -a

# Set the ESOREX recipe directory. This is not used by MicadoWISE itself,
# but can be useful if one installs ESOREX in the container.
mkdir "${HOME}/.esorex"
{
    echo "esorex.caller.recipe-dir=${CONDA_PREFIX}/lib/esopipes-plugins"
} >> "${HOME}/.esorex/esorex.rc"

# Prefetch vodslparser and ditaa.
wget https://github.com/stathissideris/ditaa/releases/download/v0.11.0/ditaa-0.11.0-standalone.jar
wget https://repo1.maven.org/maven2/org/javastro/vodsl/vodslparser/0.4.0/vodslparser-0.4.0.jar

# Put down some default configuration.
mkdir -p ~/.awe
{
    echo "[global]"
    echo "database_engine: filebased"
    echo "database_user: AWANONYMOUS"
    echo "database_password: anonymous"
    echo "data_server=localhost"
    echo "data_port=8008"
} > ~/.awe/Environment.cfg

# Cache scopesim downloads
{
    echo "alias: SIM"
    echo "properties:"
    echo "  file:"
    echo "    use_cached_downloads: True"
} > ~/.scopesim_rc.yaml

# Setup git prompt
curl https://raw.githubusercontent.com/git/git/master/contrib/completion/git-prompt.sh -o ~/git-prompt.sh
{
    echo "source ~/git-prompt.sh"
    echo "export PS1='[\u@\h \W\$(__git_ps1 \" (%s)\")]\$ '"
    echo "conda activate micadowise"
    echo "export PYTHONPATH=${HOME}/micadowise/src"
} >> ~/.bashrc

# Clone the project into /home/micado/micadowise.
# To be mounted over by the project directory on the host.
#git clone https://gitlab.astro-wise.org/micado/micadowise.git /home/micado/micadowise

# Instead, for now just move the directory.
# TODO: Decide what best to do here.
mv -v /home/micado/micadowise_install /home/micado/micadowise
