# -*- coding: utf-8 -*-
"""RecipeStep."""

import re
from typing import Type

from micado.main.commonwrapper import persistent_property
from micado.dataitems.QCParameters import QCParameters
from micado.main.DataItem import DataItem
from micado.main.Raw import Raw
from micado.main.MObject import MObject


def name_eso_from_property(prop: persistent_property, dotted: bool = False):
    """Get the ESO name of a WISE property."""
    # TODO: Integrate with persistent_property
    # TODO: Always do dotted?
    attrname_eso_docstring = prop.__doc__.split("[")[1].strip("]")
    attrname_eso_auto = (
        prop.attribute[:8] if "__" not in prop.attribute else prop.attribute
    )
    attrname_eso_auto = attrname_eso_auto.upper().replace("__", ".")
    attrname_eso = attrname_eso_docstring or attrname_eso_auto
    # TODO add this: attrname_eso = attrname_eso.strip("_")
    if dotted:
        attrname_eso = attrname_eso.replace(" ", ".")
    return attrname_eso


class RecipeStepBase:
    """A RecipeStep."""

    @classmethod
    def get_previous_step(cls: Type[MObject]):
        """Get the previous step, if any.

        TODO: Just force the property name to be 'previous'?
        """
        prev_maybe = [
            getattr(cls, pp).prop_type
            for pp in cls.get_persistent_properties()
            if type(getattr(cls, pp).prop_type) != tuple
            and issubclass(getattr(cls, pp).prop_type, RecipeStepBase)
        ]
        assert len(prev_maybe) < 2
        return prev_maybe[0] if prev_maybe else None

    @classmethod
    def get_first_step(cls):
        """Get the first step."""
        # TODO: walrus
        prev = cls
        while prev.get_previous_step() is not None:
            prev = prev.get_previous_step()
        return prev

    @classmethod
    def get_previous(cls):
        """Get the previous.

        Irrespective of whether it is another Step, a Raw, or something else.
        """
        if cls.get_previous_step() is not None:
            return cls.get_previous_step()
        elif cls.get_raw() is not None:
            return cls.get_raw()
        else:
            prevs = cls.get_other_inputdata_full()
            # TODO: Ensure that this order is actually properly preserved.
            #       Which means get_persistent_properties should be
            #       order-preserving.
            return prevs[0]

    @classmethod
    def get_raw(cls: Type[MObject]):
        """Get the raw data that is input to this class.

        TODO: Just force the property name to be 'raw'?
        """
        # TODO: Resolve DerivePeristence workaround, it is fine for now.
        #       Perhaps move to DerivePersistenceExtra if we decide to keep it.
        from micado.recipes.DerivePersistence import DerivePersistence

        if issubclass(cls, DerivePersistence):
            return Raw
        raw_maybe = [
            getattr(cls, pp).prop_type
            for pp in cls.get_persistent_properties()
            if type(getattr(cls, pp).prop_type) != tuple
            and issubclass(getattr(cls, pp).prop_type, Raw)
        ]
        assert len(raw_maybe) < 2
        return raw_maybe[0] if raw_maybe else None

    @classmethod
    def get_other_inputdata_full_attr(
        cls: Type[MObject], never_include_previous=False, include_raw=True
    ):
        """Get all the input data that is not the previous recipe or a raw.

        Also get the raw if include_raw is True.

        TODO: What about recipes that start with output from previous recipes?
              These are included if never_include_previous==False
              These are not included if never_include_previous==True
        TODO: What about class collisions?
        """
        # TODO: Integrate this into MDBObjectMeta.get_other_inputdata_full.
        cprev = (
            cls.get_previous() if never_include_previous else cls.get_previous_step()
        )
        craw = None if include_raw else cls.get_raw()
        return [
            (pp, getattr(cls, pp).prop_type)
            for pp in cls.get_persistent_properties()
            if type(getattr(cls, pp).prop_type) != tuple
            and issubclass(getattr(cls, pp).prop_type, DataItem)
            and getattr(cls, pp).prop_type not in [craw, cprev]
        ]

    @classmethod
    def get_other_inputdata_full(
        cls: Type[MObject], never_include_previous=False, include_raw=True
    ):
        """Get all the input data that is not the previous recipe or a raw.

        Also get the raw if include_raw is True.
        """
        return [
            di
            for _, di in cls.get_other_inputdata_full_attr(
                never_include_previous=never_include_previous,
                include_raw=include_raw,
            )
        ]

    @classmethod
    def get_parameter_names(cls: Type[MObject]):
        """Get the input parameter names."""
        return [param.attribute for param in cls.get_parameters()]

    @classmethod
    def get_parameters(cls: Type[MObject]):
        """Get the input parameters."""
        from micado.main.RecipeStep import RecipeStep

        # TODO: Create better way to determine what attribute is a
        #       parameter and what not.
        names_bad = {
            "object_id",
            "name",
            "filename",
            "globalname",
            "SIMPLE",
            "BITPIX",
            "NAXIS",
            # TODO: Ensure INSTNAME is not everywhere.
            "instname",
        }
        attr_cls = [
            getattr(cls, pp)
            for pp in cls.get_persistent_properties()
            if pp not in names_bad and not pp.startswith("dummy")
        ]
        return [
            prop
            for prop in attr_cls
            if type(prop.prop_type) == tuple
            or not issubclass(prop.prop_type, (DataItem, RecipeStep, QCParameters))
        ]

    @classmethod
    def name_eso(cls):
        """Name according to ESO standards. E.g. create_var_dq_layers."""
        # TODO: Refactor this duplication with Recipe.py
        # From https://stackoverflow.com/a/1176023/2097
        n1 = cls.__name__
        n2 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", n1)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", n2).lower()

    @classmethod
    def name_file_doc(cls):
        """The name to be used as a basename for documentation.

        E.g. ApplyMasterDark for ApplyMasterDark.png, ApplyMasterDark.rst,
        ApplyMasterDark.tex.
        """
        # TODO: Refactor with DataItem_extra.name_file_doc?
        return cls.__name__


class RecipeStepExtra(MObject, RecipeStepBase):
    """Persistent RecipeStepExtra.

    Must be used like this because multiple inheritance is not possible with
    persistent classes.
    """
