# -*- coding: utf-8 -*-
"""Tests for MasterDark.py."""
import os

from astropy.io import fits

from micado.config.Environment import Env
from micado.dataitems.DepersistedDark import DepersistedDark
from micado.dataitems.DepersistedExtension import DepersistedExtension
from micado.hardware.ShotNoise import ShotNoise
from micado.hardware.SummedExposure import SummedExposure
from micado.main.Img import Img
from micado.main.ObservingBlock import ObservingBlock
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.DarkCurrent import DarkCurrent
from micado.dataitems.Dark import Dark
from micado.dataitems.Dark import DarkExtension
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterDark import MasterDarkExtension
from micado.main.Imstat import Imstat
from micado.recipes.McdDetDepersist import McdDetDepersist
from micado.requirements.DPSmasterdark import DPSmasterdark
from micado.templates.MicadoImgCalDark import MicadoImgCalDark
from micado.recipes.McdDetDark import McdDetDark
from micado.test.functional.dark_test import remove_file_and_links


def mock_depersist(cls, raw):
    """Mock depersist routine.

    TODO: Put remove function.
    """
    recipe = McdDetDepersist()
    depersisted = cls(
        raw=raw,
        recipe=recipe,
    )
    depersisted.make_proto()

    return depersisted


def create_masterdark_input(simple=False, make=True, ndetectors=1):
    """Create input for MasterDark for testing.

    TODO: Replace with fixture?
    """
    # Create a MasterDark with 1 dark of 16x16 pixels when simple is set.
    n_darks = 5 if not simple else 1
    # TODO: Put 30.72 back at some point
    hw = 0.24

    mydetector = Detector(
        xhw=hw,
        yhw=hw,
    )
    mydetectorlist = [mydetector] * ndetectors

    # Set the real dark current and create a DetectorArray.
    dark_real = 100.0
    mydarkcurrent = DarkCurrent(value=dark_real)

    myshotnoise = ShotNoise(state=1)
    mydetector = DetectorArray(
        name="testdetector",
        detector_array_list=mydetectorlist,
        dark_current=mydarkcurrent,
        shot_noise=myshotnoise,
        qecurve=None,
        linearity=None,
        exposure_action=SummedExposure(),
    )

    # A template
    dark_template = MicadoImgCalDark(det__dit=1, det__ndit=1)

    # An observingblock
    dark_ob = ObservingBlock(
        name="MyDarkTest",
        instrument="MICADO",
        target="empty_sky",
        templates=[dark_template],
    )

    # Create raw Darks.
    mydarks = []
    mydepersisteddarks = []
    for _ in range(n_darks):
        # DarkExtensions
        myextensions = [
            DarkExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
        ]

        mydark = Dark(
            detector=mydetector,
            template=dark_template,
            observing_block=dark_ob,
            extensions=myextensions,
        )
        mydark.make_simulation()
        mydark.save_fits()
        mydepersisteddark = mock_depersist(DepersistedDark, mydark)
        mydepersisteddark.save_fits()

        mydarks.append(mydark)
        mydepersisteddarks.append(mydepersisteddark)

    return mydepersisteddarks, dark_real, mydarks


def create_masterdark(simple=False, make=True, ndetectors=1):
    """Create MasterDark for testing."""
    mydepersisteddarks, dark_real, mydarks = create_masterdark_input(
        simple=simple, make=make, ndetectors=ndetectors
    )

    # A recipe. The dark recipe has no parameters.
    myrecipe = McdDetDark()

    # Create MasterDark.
    mymasterdark = MasterDark(
        recipe=myrecipe,
        darks=mydepersisteddarks,
    )
    if make:
        mymasterdark.make_proto()
        mymasterdark.save_fits()
    return mydepersisteddarks, mymasterdark, dark_real, mydarks


def compare_pros(dataitem1: Img, dataitem2: Img):
    """Compare two processed Img DataItems.

    Currently works for MasterDark and MasterFlat.
    """
    assert dataitem1.fits_keywords == dataitem2.fits_keywords
    assert dataitem1.recipe == dataitem2.recipe

    ext1, ext2 = dataitem1.extensions[0], dataitem2.extensions[0]
    assert ext1.fits_keywords == ext2.fits_keywords
    assert ext1 == ext2, ext1.get_differences(ext2)

    # TODO: do something generic here
    for attr in ["darks", "flats"]:
        if hasattr(dataitem1, attr):
            for d1, d2 in zip(getattr(dataitem1, attr), getattr(dataitem2, attr)):
                assert d1.fits_keywords == d2.fits_keywords
                assert d1 == d2, d1.get_differences(d2)

    assert dataitem1 == dataitem2


def test_processing():
    """Test creation of MasterDark."""
    Env["use_python_logging"] = True
    mydarks, mymasterdark, dark_real, mydarks_raw = create_masterdark()

    # Create image statistics.
    myimstatraw = Imstat(baseframe=mydarks[0])
    myimstatraw.make()

    myimstat = Imstat(baseframe=mymasterdark)
    myimstat.make()

    # The standard deviation scales with sqrt(n_raws).
    # This check will fail if the the darks are correlated, e.g. if something
    # goes wrong with the random number generation. However, it can always
    # fail because this is a stochastic process.
    assert myimstat.std < myimstatraw.std / 1.5

    # Verify that the measured dark current is the same as the real current.
    myreq = DPSmasterdark()
    myreq.precision = 0.05
    myreq.dark_true = dark_real
    myreq.masterdarkqc = mymasterdark.qc
    myreq.make()
    assert myreq.passed, "Requirement not met!"

    myreqfail = DPSmasterdark()
    myreqfail.precision = 0.05
    myreqfail.dark_true = dark_real * 8
    myreqfail.masterdarkqc = mymasterdark.qc
    myreqfail.make()
    assert not myreqfail.passed, "Requirement met inadvertently."

    fn = mymasterdark.filename

    master2 = MasterDark(filename=fn)
    compare_pros(mymasterdark, master2)

    for dm in mydarks + mydarks_raw:
        remove_file_and_links(dm.filename)

    remove_file_and_links(mymasterdark.filename)


def test_none_mobject_fits():
    """Test an '' reference in a FITS file."""
    Env["use_python_logging"] = True

    mydarks, masterdark1, _, mydarks_raw = create_masterdark(simple=True)

    # Replace 'HIERARCH ESO REC PROCESS' = None with ''
    hdus = fits.open(os.path.realpath(masterdark1.filename), mode="update")
    hdus[0].header["HIERARCH ESO REC PROCESS"] = ""
    hdus.close()

    masterdark2 = MasterDark(filename=masterdark1.filename)
    assert masterdark1 == masterdark2

    # Cleanup.
    for dm in mydarks + mydarks_raw:
        remove_file_and_links(dm.filename)

    remove_file_and_links(masterdark1.filename)
