# -*- coding: utf-8 -*-
"""Create rst files from Python."""
from collections import OrderedDict
from pathlib import Path

from micado.dataitems.DetrendBase import DetrendBase
from micado.dataitems.DetrendHCIBase import DetrendHCIBase
from micado.dataitems.DistortionBase import DistortionBase
from micado.dataitems.DistortionElt import DistortionElt
from micado.dataitems.DistortionWam import DistortionWam
from micado.dataitems.Nonlinearity import Nonlinearity
from micado.dataitems.Persistence import Persistence
from micado.dataitems.DepersistedBase import DepersistedBase
from micado.main.RecipeStepExtra import name_eso_from_property
from micado.toolbox import rst_dataitemstable_from_python
from micado.toolbox.ditaa_from_wise import (
    PATH_DOCS_GENERATED,
    CLASSES_DATAITEMS,
    get_dataitems_derived,
)
from micado.util.esokeywords import esokeywords

PATH_TEXT_YAML = Path(__file__).resolve().parent.parent.parent / "text"

# The input can be raw data, so cannot use product.
key_input_products = "Input data"
key_output_products = "Output products"
# raw_data and processed_data are the same as input_products and
# output_products but as used in the Calibration Plan.
key_raw_data = "Instrument data"
key_processed_data = "Processed data"
key_qc_parameters = "QC parameters"
key_accuracy = "Accuracy"
key_user_parameters = "User Parameters"
key_recipeid = "DRL Recipe ID"
# Recipe name as used in the CP
key_recipe_inputto = "Input for recipe(s)"
key_recipe_outputof = "Produced by recipe"
key_name = "Name"
key_recipename = "Recipe name"
# Recipe name as used in the CP
key_recipename_cp = "Software recipe"
key_description = "Description"
key_procatg = r"\FITS{PRO.CATG}"
key_processparas = "Processing parameters"
key_docatg = r"\FITS{DO.CATG}"
key_template = "Template"
key_dpr = r"\FITS{CATG}/\FITS{TYPE}/\FITS{TECH}"
key_otherkeywords = "Other keywords"
key_oca = "OCA keywords"
key_structure = "Data item structure"
key_processing = "Processing FITS keywords"
key_mathematical_description = "Mathematical description"
key_parent_reqid = "Parent ReqID"
key_reqid = "ReqID"
key_recipe_frequency = "Recipe frequency"
key_observation_frequency = "Frequency"
key_objective = "Objective"
key_responsible = "Responsible"
key_prerequisites = "Prerequisites"
key_procedure = "Procedure"
key_remarks = "Remarks"
key_micadosetup = "MICADO Set-up"
key_calibunitsetup = "Calibration Unit setup"
key_templatename = "Observing Template"
key_templateparams = "Template parameters"
key_observation_duration = "Duration estimate [s]"
key_phase = "Phase"
# TODO: Use ObservingBlock name.
key_observingblock_name = "Observing Block"
# 611 document says purpose, used to be Objective
key_purpose = "Purpose"
# Originally "Algorithm"
key_algorithm = "DRL Functions"
key_errorconditions = "Error conditions"
key_recipetype = "Type"
key_verification = "Verification"

text_verification = "See DRL Validation and Test Plan"

dic_responsible = {
    "operations": "Science Operations",
    "user": "User",
    # TODO: 'eso' is currently only used for Persistence,
    #       should perhaps be more explicit.
    "eso": "ESO",
    "ics": "Instrumet Control Software",
    # TODO: Any is only necessary for DepersistedBase
    "any": "Any",
    "": " ",
}

dic_phase = {
    "day": "Daytime",
    "night": "Nighttime",
    # TODO: Any is only necessary for DepersistedBase
    "any": "Any",
    "": " ",
}


def keywords_from_qc(qc):
    """Convert qc to header keywords."""
    # TODO: Move to QCParameters?

    asbad = ["filename", "globalname", "object_id"]
    asqcopt = [
        a
        for a in qc.get_persistent_properties()
        if a not in asbad and not a.startswith("dummy")
    ]
    return (
        [name_eso_from_property(getattr(qc, a)) for a in asqcopt]
        if asqcopt
        else ["TODO"]
    )


def parameters_from_template(template):
    """Get comma separated list of template parameters."""
    asbad = ["filename", "globalname", "object_id"]
    astemopt = [
        a
        for a in template.get_persistent_properties()
        if a not in asbad and not a.startswith("dummy")
    ]
    as_ok = [a for a in astemopt if not a.startswith("dpr")]
    sastemplatespace = [a.upper().replace("__", " ") for a in as_ok]
    sastemplate = []
    for sa in sastemplatespace:
        # TODO: Fix yet another place with this MJDOBS MJD-OBS hack.
        if sa == "MJDOBS":
            sa = "MJD-OBS"
        # Fix INS FILT1 NAME
        sa_nonum = sa.replace("1", "").replace("2", "")
        if sa_nonum not in esokeywords:
            sa = "TPL " + sa
            assert sa in esokeywords
        sastemplate.append(sa.replace(" ", "."))

    return ", ".join(sastemplate)


def maxlen_key_value(dic: dict):
    """Get maximum width of keys and values."""
    lenk = max(len(dd) for dd in dic)
    lenv = max(max(len(cc) for cc in dd + [""]) for dd in dic.values())
    return lenk, lenv


def rst_table(dic: dict):
    """Convert dictionary to an RST table."""
    lenk, lenv = maxlen_key_value(dic)

    shline = "+-" + lenk * "-" + "-+-" + lenv * "-" + "-+"

    lstable = [shline]
    for kk, lines in dic.items():
        lines = lines or [""]
        sline = f"| {kk:{lenk}} | {lines[0]:{lenv}} |"
        lstable.append(sline)
        for line in lines[1:]:
            sline = f"| {' ':{lenk}} | {line:{lenv}} |"
            lstable.append(sline)

        lstable.append(shline)

    return "\n".join(lstable)


def rst_fieldset(dic: dict):
    """Convert dictionary to an RST fieldset."""
    # TODO: Decide whether field lists look better, depends on theme.

    lenk, lenv = maxlen_key_value(dic)
    lenk += 2
    lstable = [
        ".. bibliographic fields (which also require a transform):",
        "",
    ]
    for kk, lines in dic.items():
        sline = f"{':' + kk + ':':{lenk}} {lines[0]}"
        lstable.append(sline)
        for line in lines[1:]:
            sline = f"{' ':{lenk}} {line}"
            lstable.append(sline)

    return "\n".join(lstable)


def inputdata_from_dataitem(dataitem, allowed=None):
    """Get the input data for this datatitem.

    TODO: Refactor into dataitem itself.
    """
    recipe = dataitem.recipe.prop_type if hasattr(dataitem, "recipe") else None
    inputdata_badprogenitor = recipe.get_inputdata() if recipe else []
    inputdata = [dataitem.get_progenitor_class()] + inputdata_badprogenitor[1:]

    # TODO: Fix Persistence hack. Persistence has .raw and .raws, both are
    #       Raw, so .raws is not detected in get_other_inputdata_full().
    if dataitem == Persistence:
        inputdata = [dataitem.recipe.prop_type.process.prop_type.raws.prop_type]
    # TODO: Fix Nonlinearity hacks.
    if dataitem == Nonlinearity:
        inputdata = [dataitem.recipe.prop_type.process.prop_type.flats.prop_type]

    inputdata_multi = [get_dataitems_derived(di) for di in inputdata]
    # allowed ensures that calibration_DistortionWam.tex and ..Elt.tex
    # only have the the right input reference.
    if allowed is not None:
        inputdata_multi = [
            [a for a in inpumul if issubclass(a, allowed)]
            for inpumul in inputdata_multi
        ]

    # TODO: Fix micado_img_detrend hack.
    # This removes the HCI CORO input to the regular micado_img_detrend
    # recipe. Technically it is correct that these are there, since the
    # micado_img_detrend_hci recipe is factually the same as the
    # micado_img_detrend recipe. But they should be separate.
    if dataitem == DetrendBase:
        _, [progenitors_hci, *_] = inputdata_from_dataitem(DetrendHCIBase)
        progenitors_detrend_new = [
            di for di in inputdata_multi[0] if di not in progenitors_hci
        ]
        inputdata_multi[0] = progenitors_detrend_new

    # TODO: Fix DistortionBase hack. This can be done by making
    # DetrendFinalAstromfield and DetrendFinalWam share a base class
    # e.g. DetrendFinalDistortionBase.
    if dataitem == DistortionBase:
        inputdata_multi[0] = [
            DistortionElt.get_progenitor_class(),
            DistortionWam.get_progenitor_class(),
        ]

    # TODO: Fix Persistence hack. Persistence should just list Raw as input.
    # Actually, inputdata_multi should only be used on 'Base' classes, but
    # this does not work for Background, which should have a Base class too.
    if dataitem == Persistence:
        inputdata_multi[0] = []

    return inputdata, inputdata_multi


def dictionary_from_dataitem(dataitem):
    """Get dictionary from dataitem."""
    raw = dataitem.get_raw_class()
    template = raw.template.prop_type if raw else None
    qc = dataitem.get_qc_class()

    # Get QC parameters
    sqc = ", ".join(keywords_from_qc(qc)) if qc is not None else "None"

    # Get template parameters
    stemplatepars = (
        parameters_from_template(template) if template is not None else "None"
    )

    # Get input data
    inputdata, _ = inputdata_from_dataitem(dataitem)

    ssinput = [f"{dat.name_eso()}_" for dat in inputdata]
    sinput = ", ".join(ssinput)

    # Detrend and such should not include raw data, only calibration products
    # should do that.
    use_raw = raw.__name__ != "Science" and dataitem != Persistence

    dic1 = OrderedDict(
        [
            (key_name, dataitem.text_pro.name),
            (
                key_reqid,
                dataitem.get_requirement().name_eso()
                if dataitem.get_requirement()
                else "None",
            ),
            (
                key_parent_reqid,
                dataitem.get_requirement().get_parent_ids()
                if dataitem.get_requirement()
                else "None",
            ),
            (key_objective, dataitem.text_pro.objective),
            # Use pro responsible if set, else raw responsible.
            (
                key_responsible,
                dic_responsible[
                    dataitem.text_pro.responsible.strip()
                    or raw.text_raw.responsible.strip()
                ],
            ),
            (key_phase, dic_phase[raw.text_raw.phase.strip()] if use_raw else "N/A"),
            (key_micadosetup, raw.text_raw.micadosetup if use_raw else "N/A"),
            (key_calibunitsetup, raw.text_raw.calunitsetup if use_raw else "N/A"),
            (key_templatename, template.name_eso() if use_raw else "N/A"),
            (key_templateparams, stemplatepars if use_raw else "N/A"),
            # TODO: Do we need Observing Block name? Where to get it from?
            # (key_observingblock_name, "TODO: What is this?"),
            (key_observation_frequency, raw.text_raw.frequency if use_raw else "N/A"),
            (key_observation_duration, raw.text_raw.duration if use_raw else "N/A"),
            (key_prerequisites, raw.text_raw.prerequisites if use_raw else "N/A"),
            (key_recipe_frequency, dataitem.text_pro.frequency),
            (key_input_products, sinput),
            (key_output_products, f"{dataitem.name_eso()}_"),
            (key_qc_parameters, sqc),
            (key_recipeid, dataitem.recipe.prop_type.name_eso()),
            (key_procedure, dataitem.text_pro.procedure),
            (key_mathematical_description, dataitem.text_pro.math),
            (key_remarks, dataitem.text_pro.remarks),
            (
                key_accuracy,
                dataitem.text_pro.accuracy
                if dataitem.text_pro.accuracy != "None"
                else "TODO",
            ),
        ]
    )

    return {k: str(v).splitlines() for k, v in dic1.items()}


def table_from_dataitem(dataitem, use_field_lists=False):
    """Generate table from dataitem."""
    dic2 = dictionary_from_dataitem(dataitem)
    stable = rst_table(dic=dic2) if not use_field_lists else rst_fieldset(dic=dic2)
    prosepurpose = (
        f"""**Purpose:**

{dataitem.text_pro.prosepurpose}
"""
        if dataitem.text_pro.prosepurpose != "None"
        else ""
    )

    proseworkflow = (
        f"""**Algorithm and workflow:**

{dataitem.text_pro.proseworkflow}
"""
        if dataitem.text_pro.proseworkflow != "None"
        else ""
    )

    return stable, prosepurpose, proseworkflow, dic2


def get_keywords_table():
    """Create a table for the keywords."""
    columns = OrderedDict(
        [
            (n, [k[n] for k in esokeywords.values()])
            for n in ["name", "type", "unit", "description"]
        ]
    )
    sizes = OrderedDict([(n, max(len(s) for s in v)) for n, v in columns.items()])
    shline = " ".join("=" * size for size in sizes.values())
    sheader = " ".join(f"{n:{s}}" for n, s in sizes.items())
    sskeyword = [
        " ".join(
            f"{(v[k].replace(' ', '.') if k == 'name' else v[k]):{s}}"
            for k, s in sizes.items()
        )
        for v in esokeywords.values()
        if "____" not in v["name"]
    ]

    return "\n".join([shline, sheader, shline] + sskeyword + [shline])


def get_rst_contents_dataitem(dataitem):
    """Create contents of rst file for dataitem."""
    stable, prosepurpose, proseworkflow, _ = table_from_dataitem(dataitem)
    # TODO: Fully settle on ESO names in the documentation, also for filenames?
    fname = dataitem.__name__
    sname = dataitem.name_eso()
    sline = "^" * len(sname)
    sfigure = f"""
.. figure:: generated/{fname}.png

    Workflow diagram for {sname}
"""
    return f""".. _{sname.lower()}:

{sname}
{sline}

{prosepurpose}
{proseworkflow}

{stable}

{sfigure}
"""


def get_rst_contents_recipestep(recipestep):
    """Get rst text from Recipestep."""
    fname = recipestep.name_file_doc()
    return f"""{recipestep.name_eso()}
{'^' * len(recipestep.name_eso())}

    {recipestep.__doc__}

.. figure:: generated/{fname}.png
"""


def write_datatitem(dataitem):
    """Write a DataItem to an RST file."""
    srst = get_rst_contents_dataitem(dataitem)
    fname = dataitem.name_file_doc()
    fn_rst = PATH_DOCS_GENERATED / f"{fname}.rst"
    with open(fn_rst, "w") as f:
        f.write(srst)


def write_recipestep(recipestep):
    """Write RecipeStep to an RST file."""
    fname = recipestep.name_file_doc()
    srecipestep = get_rst_contents_recipestep(recipestep)

    fn_rst = PATH_DOCS_GENERATED / f"{fname}.rst"
    with open(fn_rst, "w") as f:
        f.write(srecipestep)


def main():
    """Convert all dataitems."""
    # TODO: Find better way to deal with DepersistedBase
    dataitems_img = [
        di for di in CLASSES_DATAITEMS if not issubclass(di, DepersistedBase)
    ]

    for dataitem in dataitems_img:
        write_datatitem(dataitem)

        recipestep = dataitem.recipe.prop_type.process.prop_type
        while recipestep is not None:
            write_recipestep(recipestep)
            recipestep = recipestep.get_previous_step()

    skeywords = get_keywords_table()
    fn_rst_keywords = PATH_DOCS_GENERATED / "keywordstable.rst"
    with open(fn_rst_keywords, "w") as f:
        f.write(skeywords)


def check_rst_contents():
    """Check whether all recipe rst files are included in recipes.rst."""
    path_rst_recipes = PATH_DOCS_GENERATED.parent / "recipes.rst"
    contents_rst_recipes = open(path_rst_recipes).readlines()
    included_in_rst = [
        line.split("/")[1].split(".")[0]
        for line in contents_rst_recipes
        if "include" in line
    ]

    dataitems_img = CLASSES_DATAITEMS

    recipe_steps = []
    for dataitem in dataitems_img:
        recipestep = dataitem.recipe.prop_type.process.prop_type
        while recipestep is not None:
            recipe_steps.append(recipestep)
            recipestep = recipestep.get_previous_step()

    extra_names = ["dataitemsraw", "dataitemsexternal", "dataitemspro"]
    item_names = [rr.__name__ for rr in dataitems_img + recipe_steps]
    items_not_in_rst = [
        r
        for r in item_names
        # TODO: Add Depersisted and Qacits
        if r not in included_in_rst
        and "Depersisted" not in r
        and "Qacits" not in r
        and "Dpsics" not in r
    ]
    items_extra_in_rst = [
        r for r in included_in_rst if r not in item_names + extra_names
    ]

    # TODO: reenable this assertion?
    if items_extra_in_rst or items_not_in_rst:
        print(
            f"""
Missing: {items_not_in_rst}
Extra: {items_extra_in_rst}
"""
        )


if __name__ == "__main__":
    main()
    rst_dataitemstable_from_python.main()
    check_rst_contents()
