"""Copy information from the LaTeX back into the YAML files.

This allows changes made directly to the DRLD/CP to be backported to
MicadoWISE.
"""

from pathlib import Path
import yaml
import dataclasses

from micado.dataitems.IllumCorr import IllumCorr
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Nonlinearity import Nonlinearity
from micado.dataitems.PhotomStdfield import PhotomStdfield
from micado.dataitems.StarcenterWaffle import StarcenterWaffle
from micado.dataitems.Staroffset import Staroffset
from micado.main.MDBObjectMeta import (
    TextRaw,
    TextPro,
    TextRecipeStep,
)

from micado.toolbox.rst_from_python import (
    PATH_DOCS_GENERATED,
    PATH_TEXT_YAML,
    CLASSES_DATAITEMS,
    key_input_products,
    key_output_products,
    key_qc_parameters,
    key_accuracy,
    key_user_parameters,
    key_recipeid,
    key_recipe_inputto,
    key_recipe_outputof,
    key_name,
    key_description,
    key_procatg,
    key_processparas,
    key_docatg,
    key_template,
    key_dpr,
    key_otherkeywords,
    key_oca,
    key_structure,
    key_processing,
    dictionary_from_dataitem,
    key_purpose,
    key_algorithm,
    key_errorconditions,
    key_recipetype,
    key_objective,
    key_procedure,
    key_templatename,
    key_recipename,
    key_responsible,
    key_phase,
    key_micadosetup,
    key_calibunitsetup,
    key_templateparams,
    key_observingblock_name,
    key_observation_duration,
    key_observation_frequency,
    key_prerequisites,
    key_processed_data,
    key_raw_data,
    key_parent_reqid,
    key_reqid,
    key_remarks,
    key_recipename_cp,
    dic_responsible,
    dic_phase,
)

keys = [
    key_input_products,
    key_output_products,
    key_qc_parameters,
    key_accuracy,
    key_user_parameters,
    key_recipeid,
    key_recipe_inputto,
    key_recipe_outputof,
    key_name,
    key_description,
    key_procatg,
    key_processparas,
    key_docatg,
    key_template,
    key_dpr,
    key_otherkeywords,
    key_oca,
    key_structure,
    key_processing,
    key_purpose,
    key_algorithm,
    key_errorconditions,
    key_recipetype,
    key_objective,
    key_procedure,
    key_templatename,
    key_recipename,
    key_responsible,
    key_phase,
    key_micadosetup,
    key_calibunitsetup,
    key_templateparams,
    key_observingblock_name,
    key_observation_duration,
    key_observation_frequency,
    key_prerequisites,
    key_processed_data,
    key_raw_data,
    key_parent_reqid,
    key_reqid,
    key_remarks,
    key_recipename_cp,
]

dic_phase_inverted = {v: k for (k, v) in dic_phase.items()}
dic_responsible_inverted = {v: k for (k, v) in dic_responsible.items()}


def dictionary_from_tex(text):
    """Parse LaTeX tex into dictionary."""
    lines_all = text.splitlines()
    lines_rows = [line for line in lines_all if "&" in line]
    dictrows = {}
    for line in lines_rows:
        key, value = [
            c.strip().strip("\\").strip().strip(":") for c in line.split("&", 1)
        ]
        if not key or "FITS" in key:
            # For multiline
            continue
        assert key in keys, key
        dictrows[key] = value
    return dictrows


def text_from_tex(tex):
    """Convert tex back to text."""
    return (
        tex.replace(r" \newline ", "\n")
        .replace(r"\%", "%")
        .replace(r"\_", "_")
        .replace(r"\newline ", "\n")
    )


def base_has_same_recipe(dataitem):
    """Check whether base class has the same recipe."""
    base = dataitem.__bases__[0]
    if hasattr(base, "recipe"):
        return base.recipe.prop_type == dataitem.recipe.prop_type
    return False


def yaml_from_dataitem(dataitem, check=True, write=False):
    """Generate YAMLs from dataitem."""

    raw = dataitem.get_raw_class()
    recipe = dataitem.recipe.prop_type

    fn_yaml_raw = PATH_TEXT_YAML / "raw" / f"{raw.__name__.lower()}.yaml"
    if fn_yaml_raw.exists():
        # Baseloader ensures that everything is a string, e.g. duration.
        dic_raw_org = yaml.load(open(fn_yaml_raw).read(), Loader=yaml.loader.BaseLoader)
        dc_raw_org = TextRaw(**dic_raw_org)
    else:
        dc_raw_org = TextRaw()

    fn_yaml_pro = PATH_TEXT_YAML / "pro" / f"{dataitem.__name__.lower()}.yaml"
    if fn_yaml_pro.exists():
        dic_pro_org = yaml.load(open(fn_yaml_pro).read(), Loader=yaml.loader.BaseLoader)
        dc_pro_org = TextPro(**dic_pro_org)
    else:
        dc_pro_org = TextPro()

    fn_tex_processed = PATH_DOCS_GENERATED / f"{dataitem.__name__}.tex"
    if fn_tex_processed.exists():
        text_processed = open(fn_tex_processed).read()
        dict_processed = dictionary_from_tex(text_processed)
    else:
        dict_processed = {}

    text_raw = open(PATH_DOCS_GENERATED / f"{raw.__name__}.tex").read()
    dict_raw = dictionary_from_tex(text_raw)

    fn_tex_calib = PATH_DOCS_GENERATED / f"calibration_{dataitem.__name__}.tex"
    if fn_tex_calib.exists():
        text_calib = open(fn_tex_calib).read()
        dict_calib = dictionary_from_tex(text_calib)
    else:
        dict_calib = {}

    text_recipe = open(PATH_DOCS_GENERATED / f"{recipe.name_eso().lower()}.tex").read()
    dict_recipe = dictionary_from_tex(text_recipe)

    dc_raw_new = TextRaw(
        responsible=dic_responsible_inverted[dict_calib[key_responsible]]
        if dict_calib
        else dc_raw_org.responsible,
        phase=dic_phase_inverted[dict_calib[key_phase]]
        if dict_calib
        else dc_raw_org.phase,
        micadosetup=text_from_tex(dict_calib[key_micadosetup])
        if dict_calib
        else dc_raw_org.micadosetup,
        calunitsetup=dict_calib[key_calibunitsetup]
        if dict_calib
        else dc_raw_org.calunitsetup,
        frequency=dict_calib[key_observation_frequency]
        if dict_calib
        else dc_raw_org.frequency,
        duration=dict_calib[key_observation_duration]
        if (dict_calib and not dc_pro_org.duration)
        else dc_raw_org.duration,
        durationper24hrs=dc_raw_org.durationper24hrs,
        prerequisites=dict_calib[key_prerequisites]
        if dict_calib
        else dc_raw_org.prerequisites,
        keywordsprocessing=dc_raw_org.keywordsprocessing,
    )
    if fn_yaml_raw.exists() and check:
        for field in dataclasses.fields(dc_raw_org):
            v_org = dataclasses.asdict(dc_raw_org)[field.name]
            v_new = dataclasses.asdict(dc_raw_new)[field.name]
            if v_org == "null" and v_new == "None":
                continue
            assert v_org == v_new, (field.name, v_org, v_new)
    if fn_yaml_raw.exists() and write:
        with open(fn_yaml_raw, "w") as f:
            yaml.dump(dataclasses.asdict(dc_raw_new), f)

    if dict_calib:
        # E.g. overrules DRLD for IllumCorr
        s_remarks = text_from_tex(dict_calib[key_remarks])
    elif base_has_same_recipe(dataitem):
        # This never shows up in the DRLD
        s_remarks = dc_pro_org.remarks
    else:
        # Get from DRLD
        s_remarks = text_from_tex(dict_recipe[key_remarks])

    dc_pro_new = TextPro(
        # Cannot use dict_processed here because that has hyperref.
        name=dict_calib[key_name] if dict_calib else dc_pro_org.name,
        objective=dict_calib[key_objective] if dict_calib else dc_pro_org.objective,
        # frequency from dict_processed is not used in the tex
        # frequency=dict_calib[key_observation_frequency],
        frequency=dc_pro_org.frequency,
        # duration=dict_calib[key_observation_duration],
        duration=dc_pro_org.duration,
        # procedure isn't either?
        # procedure=text_from_tex(dict_recipe[key_procedure]),
        procedure=dc_pro_org.procedure,
        # math not used either?
        math=dc_pro_org.math,
        remarks=s_remarks,
        accuracy=text_from_tex(dict_calib[key_accuracy])
        if dict_calib
        else dc_pro_org.accuracy,
        responsible=dic_responsible_inverted[dict_calib[key_responsible]]
        if dict_calib
        else dc_pro_org.responsible,
        prerequisites=dict_calib[key_prerequisites]
        if dict_calib
        else dc_pro_org.prerequisites,
        prosepurpose=dc_pro_org.prosepurpose,
        proseworkflow=dc_pro_org.proseworkflow,
        # TODO: automate processing keywords
        keywordsprocessing=dc_pro_org.keywordsprocessing,
    )
    if fn_yaml_pro.exists() and check:
        for field in dataclasses.fields(dc_pro_org):
            if field.name in ["responsible"]:
                continue
            if field.name == "prerequisites" and dataitem in [
                IllumCorr,
                MasterFlat,
                Nonlinearity,
                PhotomStdfield,
                StarcenterWaffle,
                Staroffset,
            ]:
                continue
            if field.name == "accuracy" and dataitem in [
                MasterDark,
                MasterFlat,
                StarcenterWaffle,
            ]:
                continue
            v_org = dataclasses.asdict(dc_pro_org)[field.name]
            v_new = dataclasses.asdict(dc_pro_new)[field.name]
            if v_org == "null" and v_new == "None":
                continue
            if isinstance(v_org, str):
                v_org = "\n".join(
                    l.strip() for l in v_org.strip().strip(".").splitlines()
                )
                v_new = "\n".join(
                    l.strip() for l in v_new.strip().strip(".").splitlines()
                )
            assert v_org == v_new, (field.name, v_org, v_new)
        # assert dc_pro_new == dc_pro_org
    if fn_yaml_pro.exists() and write:
        with open(fn_yaml_pro, "w") as f:
            yaml.dump(dataclasses.asdict(dc_pro_new), f)


if __name__ == "__main__":
    for dataitem in CLASSES_DATAITEMS:
        yaml_from_dataitem(dataitem, check=False, write=True)
